package org.sc3d.apt.family.v1;

import java.io.*;
import java.util.*;

public class MakeHTML {
  /** Don't want any instances of this class. */
  private MakeHTML() {}
  
  /** The argument type of 'write()'. */
  public static class LinkStyle {
    /** Returns HTML for a link to a page about 'person', or just their name if there is no such page. The default implementation makes a link to 'person.key+".html"'. */
    public String link(Person person) {
      return "<a href=\""+person.key+".html\">"+person.names+"</a>";
    }
  }
  
  /** Writes an HTML page about 'person' to 'out'.
   * @param person the subject of the page.
   * @param out the PrintStream to which to write (e.g. 'System.out').
   * @param linkStyle the algorithm for generating links to other people.
   */
  public static void write(
    final Person person,
    final PrintStream out,
    final LinkStyle ls
  ) throws IOException {
    out.println(
      "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" "+
      "http://www.w3.org/TR/REC-html40/loose.dtd>"
    );
    out.println(
      "<!-- This page was automatically generated from a family-tree file. If "+
      "you edit the page by hand, you risk losing your changes next time the "+
      "page is regenerated. -->"
    );
    out.println("<html>");
    out.println(" <head>");
    out.println(
      "  <!-- Generated by org.sc3d.apt.family.v1.MakeHTML on "+new Date()+
      ". -->"
    );
    out.println("  <title>Family tree for "+person.names+"</title>");
    out.println(" </head><body>");
    out.println("  <h1>Family tree of "+person.names+"</h1>");
    // Print the table of information about 'person'.
    out.println("  <table align=\"center\" border=\"1\">");
    out.print("   <tr><td colspan=\"2\">");
    if (person.titles!=null) out.print(person.titles+" ");
    out.print(person.names==null ? "Name not known" : person.names);
    if (person.qualifications!=null) out.print(" "+person.qualifications);
    out.println("</td></tr>");
    if (person.sex!=null)
      out.println("   <tr><td>Sex</td><td>"+person.sex+"</td></tr>");
    if (person.profession!=null) out.println(
      "   <tr><td>Profession</td><td>"+person.profession+
      "</td></tr>"
    );
    if (person.birth!=null)
      out.println("   <tr><td>Born</td><td>"+person.birth+"</td></tr>");
    if (person.death!=null)
      out.println("   <tr><td>Died</td><td>"+person.death+"</td></tr>");
    if (person.parents!=null) {
      final Person.Marriage parents = person.parents;
      out.println(
        "   <tr><td>Father</td><td>"+ls.link(parents.father)+"</td></tr>"
      );
      out.println(
        "   <tr><td>Mother</td><td>"+ls.link(parents.mother)+"</td></tr>"
      );
    }
    for (int i=0; i<person.marriages.length; i++) {
      final Person.Marriage m = person.marriages[i];
      final Person spouse = person==m.father ? m.mother : m.father;
      out.println(
        "   <tr><td rowspan=\"2\" valign=\"top\">Married</td><td>"+
        ls.link(spouse)+"</td></tr><tr><td>"+m.wedding+"</td></tr>"
      );
      if (m.children.length>0) {
        out.println(
          "   <tr><td rowspan=\""+m.children.length+"\" valign=\"top\">"+
          "Children</td><td>"+ls.link(m.children[0])+"</td></tr>"
        );
        for (int j=1; j<m.children.length; j++)
          out.println("   <tr><td>"+ls.link(m.children[j])+"</td></tr>");
      }
    }
    out.println("  </table>");
    for (int i=0; i<person.notes.length; i++)
      out.println("  <p>"+person.notes[i]+"</p>");
    out.println("  <h2>Ancestors</h2>");
    out.println("  <ul>");
    MakeHTML.writeAncestors(person, out, ls, "  ");
    out.println("  </ul>");
    out.println("  <h2>Descendents</h2>");
    out.println("  <ul>");
    MakeHTML.writeDescendants(person, out, ls, "  ");
    out.println("  </ul>");
    out.println(" </body>");
    out.println("</html>");
  }
  
  /** Writes HTML for the ancestor list of 'person', starting with "<li>".
   * @param person the root of the list.
   * @param out the PrintStream to which to write.
   * @param ls the algorithm used to make links to pages about other people.
   * @param indent a String that will be prepended to every line.
   */
  public static void writeAncestors(
    final Person person,
    final PrintStream out,
    final LinkStyle ls,
    final String indent
  ) throws IOException {
    out.println(indent+"<li>"+ls.link(person));
    String newIndent = indent + " ";
    if (person.parents!=null) {
      out.println(newIndent+"<ul>");
      MakeHTML.writeAncestors(person.parents.father, out, ls, newIndent);
      MakeHTML.writeAncestors(person.parents.mother, out, ls, newIndent);
      out.println(newIndent+"</ul>");
    }
  }
  
  /** Writes HTML for the descendant list of 'person', starting with "<li>".
   * @param person the root of the list.
   * @param out the PrintStream to which to write.
   * @param ls the algorithm used to make links to pages about other people.
   * @param indent a String that will be prepended to every line.
   */
  public static void writeDescendants(
    final Person person,
    final PrintStream out,
    final LinkStyle ls,
    final String indent
  ) throws IOException {
    out.println(indent+"<li>"+ls.link(person));
    String newIndent = indent + " ";
    out.println(newIndent+"<ul>");
    for (int i=0; i<person.marriages.length; i++) {
      final Person.Marriage m = person.marriages[i];
      for (int j=0; j<m.children.length; j++) {
        MakeHTML.writeDescendants(m.children[j], out, ls, newIndent);
      }
    }
    out.println(newIndent+"</ul>");
  }
  
  /** Arguments are interpreted as filenames of family-tree files. All are read, then a page is written to the directory 'tree' about each person. */
  public static void main(String[] args) throws IOException {
    if (args.length==0) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.family.v1.MakeHTML <file> <file> ..."
    );
    final LinkStyle ls = new LinkStyle();
    // Read the input.
    for (int i=0; i<args.length; i++) Person.read(new FileReader(args[i]));
    // Write a page for each Person.
    for (Iterator it=Person.PEOPLE.keySet().iterator(); it.hasNext(); ) {
      final Person person = (Person)(Person.PEOPLE.get(it.next()));
      final PrintStream out =
        new PrintStream(new FileOutputStream("tree/"+person.key+".html"));
      MakeHTML.write(person, out, ls);
    }
    // Write an index page.
    final PrintStream out =
      new PrintStream(new FileOutputStream("tree/index.html"));
    out.println(
      "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" "+
      "http://www.w3.org/TR/REC-html40/loose.dtd>"
    );
    out.println(
      "<!-- This page was automatically generated from a family-tree file. If "+
      "you edit the page by hand, you risk losing your changes next time the "+
      "page is regenerated. -->"
    );
    out.println("<html>");
    out.println(" <head>");
    out.println(
      "  <!-- Generated by org.sc3d.apt.family.v1.MakeHTML on "+new Date()+
      ". -->"
    );
    out.println("  <title>Family tree</title>");
    out.println(" </head><body>");
    out.println("  <h1>Family tree</h1>");
    out.println("  <p>");
    for (
      Iterator it=new TreeSet(Person.PEOPLE.keySet()).iterator();
      it.hasNext();
    ) {
      final Person person = (Person)(Person.PEOPLE.get(it.next()));
      out.println(ls.link(person)+".");
    }
    out.println("  </p>");
    out.println(" </body>");
    out.println("</html>");
  }
}
