package org.sc3d.apt.family.v1;

import java.io.*;
import java.util.*;

/** Represents a person in the family tree. Every person has a unique key. */
public class Person {
  /** Constructs a Person.
   * @param key the unique key of this Person.
   * @throws IllegalArgumentException if the key is not unique.
   */
  public Person(String key) {
    if (PEOPLE.containsKey(key)) throw new IllegalArgumentException(
      "There is already a Person "+key+" so you can't create another"
    );
    PEOPLE.put(key, this);
    this.key = key;
    this.marriages = new Marriage[0];
    this.notes = new String[0];
  }
  
  /* Static. */
  
  /** A Map from all unique keys used so far to the corresponding Persons. */
  public static final Map PEOPLE = new HashMap();
  
  /** Writes a complete family-tree file to 'out'. */
  public static void write(Writer out) throws IOException {
    final HashSet ms = new HashSet(); // All Marriages found.
    final LinkedList mlist = new LinkedList(); // The order of Marriages.
    for (Iterator it=new TreeSet(PEOPLE.keySet()).iterator(); it.hasNext(); ) {
      final Person p = (Person)(PEOPLE.get(it.next()));
      out.write(p.toString()); out.write('\n');
      for (int i=0; i<p.notes.length; i++) {
        out.write("NOTE "); out.write(p.key);
        out.write(' '); out.write(p.notes[i]); out.write('\n');
      }
      if (p.parents!=null && !ms.contains(p.parents)) {
        ms.add(p.parents); mlist.add(p.parents);
      }
      for (int i=0; i<p.marriages.length; i++) ms.add(p.marriages[i]);
    }
    for (Iterator it = mlist.iterator(); it.hasNext(); ) {
      out.write(it.next().toString()); out.write('\n');
    }
  }
  
  /** Reads a complete family-tree file from 'in'. */
  public static void read(Reader in) throws IOException {
    final BufferedReader br = new BufferedReader(in);
    final Set pNotDone = new HashSet();
    while (true) {
      String line = br.readLine();
      if (line==null) break;
      StringTokenizer st = new StringTokenizer(line);
      if (st.hasMoreTokens()) {
        String t = st.nextToken();
        if (t.equals("#")) {
          // Do nothing.
        } else if (t.equals("PERSON")) {
          t = st.nextToken();
          System.out.println("PERSON "+t);
          final Person p = pNotDone.contains(t)
            ? (Person)PEOPLE.get(t)
            : new Person(t);
          pNotDone.remove(t);
          if (st.hasMoreTokens()) {
            // Sex.
            t = st.nextToken();
            if (t.equals("?")) p.sex = null;
            else if (t.equals("M")) p.sex = Sex.MALE;
            else if (t.equals("F")) p.sex = Sex.FEMALE;
            else throw new RuntimeException("Sex should be ?, M or F");
          }
          // Names and other things.
          char field = 'n';
          while (st.hasMoreTokens()) {
            t = st.nextToken();
            if (t.equals("BORN")) {
              field = '?';
              p.birth = new Event(st.nextToken(), st.nextToken());
            } else if (t.equals("DIED")) {
              field = '?';
              p.death = new Event(st.nextToken(), st.nextToken());
            } else if (t.equals("PROFESSION")) {
              field = 'p';
            } else if (t.equals("TITLES")) {
              field = 't';
            } else if (t.equals("QUALIFICATIONS")) {
              field = 'q';
            } else switch (field) {
              case '?':
                throw new RuntimeException(
                  "Expected BORN, DIED, PROFESSION, TITLES or QUALIFICATIONS "+
                  "but found "+t
                );
              case 'n':
                p.names = p.names==null ? t : p.names+' '+t;
                break;
              case 'p':
                p.profession = p.profession==null ? t : p.profession+' '+t;
                break;
              case 't':
                p.titles = p.titles==null ? t : p.titles+' '+t;
                break;
              case 'q':
                p.qualifications =
                  p.qualifications==null ? t : p.qualifications+' '+t;
                break;
            }
          }
        } else if (t.equals("NOTE")) {
          t = st.nextToken();
          System.out.println("NOTE "+t);
          Person p = (Person)(PEOPLE.get(t));
          if (p==null) { pNotDone.add(t); p = new Person(t); }
          StringBuffer sb = new StringBuffer(st.nextToken());
          while (st.hasMoreTokens()) sb.append(' ').append(st.nextToken());
          p.addNote(sb.toString());
        } else if (t.equals("MARRIAGE")) {
          String ft = st.nextToken(), mt = st.nextToken();
          System.out.println("MARRIAGE "+ft+" "+mt);
          Person fp = (Person)(PEOPLE.get(ft));
          if (fp==null) { pNotDone.add(ft); fp = new Person(ft); }
          Person mp = (Person)(PEOPLE.get(mt));
          if (mp==null) { pNotDone.add(mt); mp = new Person(mt); }
          final Marriage m = new Marriage(fp, mp);
          if (st.hasMoreTokens()) m.wedding =
            new Event(st.nextToken(), st.nextToken());
          if (st.hasMoreTokens()) {
            t = st.nextToken();
            if (!t.equals("CHILDREN")) throw new RuntimeException(
              "Expected CHILDREN but found "+t
            );
            while (st.hasMoreTokens()) {
              t = st.nextToken();
              Person p = (Person)(PEOPLE.get(t));
              if (p==null) { pNotDone.add(t); p = new Person(t); }
              m.addChild(p);
            }
          }
        } else throw new RuntimeException(
          "Don't know what to do with lines beginning "+t
        );
      }
    }
    if (pNotDone.size()!=0) {
      StringBuffer sb = new StringBuffer("PERSON lines needed for:");
      for (Iterator it=pNotDone.iterator(); it.hasNext(); )
        sb.append(' ').append(it.next());
      throw new RuntimeException(sb.toString());
    }
  }
  
  /* New API. */

  /** The unique key of this person, as passed to the constructor. Example: "HP1983" */
  public final String key;
  
  /** The sex of this person, of 'null' if information is not available. */
  public Sex sex;
  
  /** The full title of this person, or 'null' if information is not available. Example: "Sir" */
  public String titles;

  /** The full name of this person, or 'null' if information is not available. Example: "Harry Potter" */
  public String names;
    
  /** The qualifications of this person, or 'null' if information is not available. Example: "QC" */
  public String qualifications;
  
  /** The profession of this Person, or 'null' if information is not available. Example: "farmer" */
  public String profession;

  /** The birth of this Person, or 'null' if information is not available. */
  public Event birth;
  
  /** The death of this Person, or 'null' if information is not available. */
  public Event death;
  
  /** This Person's parents, or 'null' if information is not available. */
  public Marriage parents;
  
  /** This Person's marriages. */
  public Marriage[] marriages;
  
  /** Adds a Marriage to 'marriages'. */
  public void addMarriage(Marriage m) {
    final Marriage[] old = this.marriages;
    this.marriages = new Marriage[old.length+1];
    System.arraycopy(old, 0, this.marriages, 0, old.length);
    this.marriages[old.length] = m;
  }
  
  public String[] notes;
  
  /** Adds a note to 'notes'. */
  public void addNote(String n) {
    final String[] old = this.notes;
    this.notes = new String[old.length+1];
    System.arraycopy(old, 0, this.notes, 0, old.length);
    this.notes[old.length] = n;
  }
  
  /* Inner classes. */
  
  /** Represents a marriage. */
  public static final class Marriage {
    /** Constructs a Marriage, and adds it to both parents. */
    public Marriage(Person father, Person mother) {
      this.father = father; this.mother = mother;
      father.addMarriage(this); mother.addMarriage(this);
      this.children = new Person[0];
    }
    
    /** A parent, as passed to the constructor. */
    public final Person father, mother;
    
    /** The wedding, or 'null' if no information is available. */
    public Event wedding;
    
    /** Children from this marriage. */
    public Person[] children;
    
    /** Adds a child to 'children', and adds 'father' and 'mother' to 'child'.
     * @throws IllegalArgumentException if 'child' already has parents.
     */
    public void addChild(Person p) {
      if (p.parents!=null) throw new IllegalArgumentException(
        p.key+" already has parents."
      );
      p.parents = this;
      final Person[] old = this.children;
      this.children = new Person[old.length+1];
      System.arraycopy(old, 0, this.children, 0, old.length);
      this.children[old.length] = p;
    }
    
    /** Returns a "MARRIAGE" line for this Marriage. */
    public String toString() {
      StringBuffer sb = new StringBuffer("MARRIAGE");
      sb.append(' ').append(this.father.key);
      sb.append(' ').append(this.mother.key);
      if (this.wedding!=null) sb.append(' ').append(this.wedding);
      sb.append(" CHILDREN");
      for (int i=0; i<this.children.length; i++)
        sb.append(' ').append(this.children[i].key);
      return sb.toString();
    }
  }
  
  /* Override things in Object. */
  
  /** Returns a "PERSON" line for this Person. */
  public String toString() {
    StringBuffer sb = new StringBuffer("PERSON");
    sb.append(' ').append(this.key);
    sb.append(' ').append(this.sex==null ? "?" : this.sex.toString());
    if (this.names!=null) sb.append(' ').append(this.names);
    if (this.birth!=null) sb.append(" BORN ").append(this.birth);
    if (this.death!=null) sb.append(" DIED ").append(this.death);
    if (this.titles!=null) sb.append(" TITLES ").append(this.titles);
    if (this.qualifications!=null)
      sb.append(" QUALIFICATIONS ").append(this.qualifications);
    if (this.profession!=null)
      sb.append(" PROFESSION ").append(this.profession);
    return sb.toString();
  }
  
  /* Test code. */
  
  /** args[0] is interpreted as the filename of a family-tree file. The file is read in and then written out again. */
  public static void main(String[] args) throws IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.family.v1.Person <filename>"
    );
    final String filename = args[0];
    final Reader in = new FileReader(filename);
    Person.read(in);
    System.out.println("\n\nFile successfully read.\n\n");
    final Writer out = new OutputStreamWriter(System.out);
    Person.write(out);
    out.flush();
  }
}
